//
// tournament - creation of playing schedules for tournaments
// Copyright (C) 2011  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

/// \file
/// \brief Main file.

#include <cassert>
#include <cmath>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <limits>
#include <list>
#include <sstream>
#include <vector>

#ifdef HAVE_OPENMP
#include <omp.h>
#endif // HAVE_OPENMP

#include "timer.h"

/// Number of male players
int K;
/// Number of female players
int L;
/// Total Number of players
int KpL;
/// Number of rounds
int M;
/// Number of courts
int N;
/// Volatility of the simulated annealing
float V;
/// Lower bound of a tournament
int g_lower_bound;

/// Helper class that supports output with padding.
///
/// Its output operator invokes the output operator of the base class of type T.
/// If necessary, the output is then padded to the given width.
template<class T>
class Padding : public T
{
public:
    /// Constructor.
    ///
    /// Wraps the instance \p t such that its output operator writes at least
    /// \p width characters.
    Padding( const T& t, size_t width): T( t), m_width( width) { };

    /// The desired width.
    const size_t m_width;
};

/// Output operator
template<class T>
std::ostream& operator<<( std::ostream& stream, const Padding<T>& p)
{
    std::ostringstream s;
    s << * static_cast<const T*>( &p);
    for( size_t i = s.str().size(); i < p.m_width; ++i)
        s << " ";
    stream << s.str();
    return stream;
}

/// Represents a player.
///
/// Players are identified by an ID. The IDs 0..#K-1 indicate the male players,
/// the IDs #K..#K+#L-1 indicate the female players. The output operator uses
/// letters 'A', 'B', ... for male players, and numbers 0, 1, ... for female
/// players.
class Player
{
public:
    /// Constructor.
    ///
    /// Constructs a player with ID 0.
    Player() : m_id( 0) { }

    /// Constructor.
    ///
    /// Constructs a player with ID \p id.
    Player( int id) : m_id( id) { }
    
    /// Sets the player ID.
    ///
    /// \param id   The new player ID.
    void set_id( int id) { assert( id >= 0); assert( id < KpL); m_id = id; }

    /// Returns the player ID.
    int get_id() const { return m_id; }
    
    /// Indicates whether the player is male.
    bool is_male() const { return m_id < K; }
    
    /// Indicates whether the players is female.
    bool is_female() const { return m_id >= K; }
    
    /// Equality operator.
    bool operator==( const Player& rhs) const { return m_id == rhs.m_id; }
    
    /// Inequality operator.
    bool operator!=( const Player& rhs) const { return m_id != rhs.m_id; }

private:
    /// The player ID.
    int m_id;
};

/// Output operator
std::ostream& operator<<( std::ostream& stream, const Player& p)
{
    int id = p.get_id();
    if( p.is_male())
        stream << static_cast<char>( 'A' + id);
    else
        stream << id - (K-1);
    return stream;
}
        
/// Represents a match.
///
/// A match consists of four players, inditified by indices 0 to 3. The first
/// team consists of players 0 and 1, the second team consists of players 2 and
/// 3. The players 0 and 2 are male, the players 1 and 3 are female.
class Match
{
public:
    /// Constructor.
    ///
    /// Constructs a match with players 0 and K against players 1 and K+1.
    Match()
    {
        m_player[0] = Player( 0);
        m_player[1] = Player( K);
        m_player[2] = Player( 1);
        m_player[3] = Player( K+1);
    }
    
    /// Sets a player of the match.
    ///
    /// \param index   The player to change.
    /// \param player  The new player.
    void set_player( int index, const Player& player)
    {
        assert( index >= 0);
        assert( index < 4);
        m_player[index] = player;
    }
    
    /// Returns a player of the match
    ///
    /// \param index   The player to return.
    /// \return        The player with index \p index of the match.
    const Player& get_player( int index) const
    {
        assert( index >= 0);
        assert( index < 4);
        return m_player[index];
    }
    
    /// Indicates whether the match is valid.
    ///
    /// \return   \c true if the male players are different and the female
    ///           players are different
    bool is_valid()
    {
        assert( m_player[0].is_male());
        assert( m_player[1].is_female());
        assert( m_player[2].is_male());
        assert( m_player[3].is_female());

        return (m_player[0] != m_player[2]) && (m_player[1] != m_player[3]);
    }

private:
    /// The players of the match.
    Player m_player[4];
};

/// Output operator
std::ostream& operator<<( std::ostream& stream, const Match& m)
{
    stream << m.get_player( 0) << Padding<Player>( m.get_player( 1), L>9 ? 2 : 1) << " : "
           << m.get_player( 2) << Padding<Player>( m.get_player( 3), L>9 ? 2 : 1);
    return stream;
}

/// Represents a tournament.
///
/// A tournament consists of rounds. In each round there is exactly one match
/// per court.
///
/// Uses the global variables \p M and \p N.
class Tournament
{
public:
    /// Constructor.
    ///
    /// Creates a tournament with #M rounds and #N courts.
    Tournament()
    {
        m_rounds = M;
        m_courts = N;
        m_matches = new Match[m_rounds * m_courts];

        int k = 0;
        int l = 0;
        for( int i = 0; i < m_rounds * m_courts; ++i) {
            m_matches[i].set_player( 0, 0 + k++);
            if( k >= K) k = 0;
            m_matches[i].set_player( 1, K + l++);
            if( l >= L) l = 0;
            m_matches[i].set_player( 2, 0 + k++);
            if( k >= K) k = 0;
            m_matches[i].set_player( 3, K + l++);
            if( l >= L) l = 0;
        }
    }
    
    /// Copy constructor.
    Tournament( const Tournament& other)
    {
        m_rounds = other.m_rounds;
        m_courts = other.m_courts;
        m_matches = new Match[m_rounds * m_courts];
        std::copy( other.m_matches,
            other.m_matches + other.m_rounds * other.m_courts, m_matches);
    }
    
    /// Assignment operator.
    ///
    /// Requires that \c *this and \c rhs agree on the number of rounds and courts.
    Tournament& operator=( const Tournament& rhs)
    {
        assert( m_rounds == rhs.m_rounds);
        assert( m_courts == rhs.m_courts);

        if( this == &rhs)
            return *this;
        
        std::copy( rhs.m_matches,
            rhs.m_matches + rhs.m_rounds * rhs.m_courts, m_matches);
        return *this;
    }

    /// Destructor
    ~Tournament()
    {
        delete[] m_matches;
        m_matches = 0;
    }

    /// Returns the number of rounds.
    int get_rounds() const { return m_rounds; }
    
    /// Returns the number of courts.
    int get_courts() const { return m_courts; }

    /// Returns a match.
    ///
    /// \param round   The round of the match to return.
    /// \param court   The court of the match to return.
    /// \return        The match on court \p court in round \p round.
    const Match& get_match( int round, int court) const
    {   
        return m_matches[ round * m_courts + court];
    }
    
    /// Returns a match.
    ///
    /// \param round   The round of the match to return.
    /// \param court   The court of the match to return.
    /// \return        The match on court \p court in round \p round.
    Match& get_match( int round, int court)
    {   
        return m_matches[ round * m_courts + court];
    }

private:
    /// The number of rounds.
    int m_rounds;

    /// The number of courts.
    int m_courts;

    /// The matches.
    Match* m_matches;
};

/// Output operator
std::ostream& operator<<( std::ostream& stream, const Tournament& t)
{
    stream << "          ";
    if( M > 9) stream << " ";
    if( L > 9) stream << " ";
    for( int j = 0; j < t.get_courts(); ++j) {
         stream << "court " << j+1 << "   ";
         if( L > 9) stream << "  ";
    }
    stream << std::endl;

    for( int i = 0; i < t.get_rounds(); ++i) {
        stream << "round " << i+1 << "   ";
        if( M > 9 && i < 9) stream << " ";
        for( int j = 0; j < t.get_courts(); ++j)
            stream << t.get_match( i, j) << "   ";
        stream << std::endl;
    }
    return stream;
}
        
/// Compute penalties for every player that plays mutiple times in the same
/// round.
///
/// \param t        The tournament for which to compute the penalty.
/// \param dump     If \c true, dump a message for each penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param weight   The weight for this penalty.
/// \return         The computed penalty.
int penalty_multiple_per_round( const Tournament& t, int* buffer, bool dump, int weight)
{
    int total_penalty = 0;
    bool dumped = false;
    
    for( int i = 0; i < t.get_rounds(); ++i) {

        for( int p = 0; p < KpL; ++p)
            buffer[p] = 0;

        // For all p in [0,KpL): buffer[p] indicates how often player p
        // plays in round i.
        for( int j = 0; j < t.get_courts(); ++j) {
            const Match& match = t.get_match( i, j);
            buffer[ match.get_player( 0).get_id() ]++;
            buffer[ match.get_player( 1).get_id() ]++;
            buffer[ match.get_player( 2).get_id() ]++;
            buffer[ match.get_player( 3).get_id() ]++;
        }

        // Let c indicate how ofter player p plays in round i.
        // If c > 1 add penalty c-1.
        for( int p = 0; p < KpL; ++p) {
            int c = buffer[p];
            if( c <= 1)
                 continue;
            int penalty = weight * (c-1);
            total_penalty += penalty;
            if( dump) {
                std::cout << "player " << Player( p) << " plays "
                          << c << " times in round " << i
                          << " (penalty: " << penalty << ")" << std::endl;
                dumped = true;
            }
        }
    }

    if( dumped)
        std::cout << std::endl;
    return total_penalty;
}

/// Compute penalties for every pair of players that play together multiple times.
///
/// \param t        The tournament for which to compute the penalty.
/// \param dump     If \c true, dump a message for each penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param weight   The weight for this penalty.
/// \return         The computed penalty.
int penalty_same_teammate( const Tournament& t, int* buffer, bool dump, int weight)
{
    int total_penalty = 0;
    bool dumped = false;
    for( int p = 0; p < KpL*KpL; ++p)
        buffer[p] = 0;

    // For all p, q in [0,KpL): buffer[ p*KpL + q ] indicates how often
    // player p and q play together.
    for( int i = 0; i < t.get_rounds(); ++i) {
        for( int j = 0; j < t.get_courts(); ++j) {
            const Match& match = t.get_match( i, j);
            int p0 = match.get_player( 0).get_id();
            int p1 = match.get_player( 1).get_id();
            int p2 = match.get_player( 2).get_id();
            int p3 = match.get_player( 3).get_id();
            buffer[ p0*KpL + p1 ]++;
            buffer[ p1*KpL + p0 ]++;
            buffer[ p2*KpL + p3 ]++;
            buffer[ p3*KpL + p2 ]++;
        }
    }

    // Let c indicate how ofter player p and q play together.
    // If c > 1 add penalty c*c - 1.
    for( int p = 0; p < KpL; ++p) {
        for( int q = 0; q < KpL; ++q) {
            int c = buffer[ p*KpL + q ];
            assert( (p != q) || (c == 0));
            if( c <= 1)
                continue;
            int penalty = weight * (c-1) * (c-1);
            total_penalty += penalty;
            if( dump && p<q) {
                std::cout << "players " << Player( p) << " and " << Player( q)
                          << " play " << c << " times as a team"
                          << " (penalty: " << penalty << ")" << std::endl;
                dumped = true;
            }
        }
    }

    if( dumped)
        std::cout << std::endl;
    assert( total_penalty % (2 * weight) == 0);
    return total_penalty/2;
}

/// Compute penalties for every pair of players that play against each other
/// multiple times.
///
/// \param t        The tournament for which to compute the penalty.
/// \param dump     If \c true, dump a message for each penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param weight   The weight for this penalty.
/// \return         The computed penalty.
int penalty_same_opponent( const Tournament& t, int* buffer, bool dump, int weight)
{
    int total_penalty = 0;
    bool dumped = false;
    for( int p = 0; p < KpL*KpL; ++p)
        buffer[p] = 0;

    // For all p, q in [0,KpL): buffer[ p*KpL + q ] indicates how often
    // player p and q play against each other.
    for( int i = 0; i < t.get_rounds(); ++i) {
        for( int j = 0; j < t.get_courts(); ++j) {
            const Match& match = t.get_match( i, j);
            int p0 = match.get_player( 0).get_id();
            int p1 = match.get_player( 1).get_id();
            int p2 = match.get_player( 2).get_id();
            int p3 = match.get_player( 3).get_id();
            buffer[ p0*KpL + p2 ]++;
            buffer[ p0*KpL + p3 ]++;
            buffer[ p1*KpL + p2 ]++;
            buffer[ p1*KpL + p3 ]++;
            buffer[ p2*KpL + p0 ]++;
            buffer[ p2*KpL + p1 ]++;
            buffer[ p3*KpL + p0 ]++;
            buffer[ p3*KpL + p1 ]++;
        }
    }

    // Let c indicate how ofter player p and q play against each other.
    // If c > 1 add penalty c*c - 1.
    for( int p = 0; p < KpL; ++p) {
        for( int q = 0; q < KpL; ++q) {
            int c = buffer[p*KpL + q];
            assert( (p != q) || (c == 0));
            if( c <= 1)
                continue;
            int penalty = weight * (c-1) * (c-1);
            total_penalty += penalty;
            if( dump && p<q) {
                std::cout << "players " << Player( p) << " and " << Player( q)
                          << " play " << c << " times against each other"
                          << " (penalty: " << penalty << ")" << std::endl;
                dumped = true;
            }
        }
    }

    if( dumped)
        std::cout << std::endl;
    assert( total_penalty % (2 * weight) == 0);
    return total_penalty/2;
}

/// Compute penalties for subsequent matches without break.
///
/// \param t        The tournament for which to compute the penalty.
/// \param dump     If \c true, dump a message for each penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param weight   The weight for this penalty.
/// \return         The computed penalty.
int penalty_subsequent_matches( const Tournament& t, int* buffer, bool dump, int weight)
{
    int total_penalty = 0;
    bool dumped = false;
    for( int p = 0; p < KpL; ++p)
        buffer[ p ] = 0;
    for( int p = 0; p < KpL; ++p)
        buffer[ KpL + p ] = -1;

    // For all p in [0,KpL): buffer[ p ] indicates the maximum number of subsequent
    // matches of player p (so far), buffer[ KpL + p ] indicates the last round
    // where player p had a break, and buffer [ 2*KpL + p ] indicates whether
    // player p has a match in this round.
    for( int i = 0; i < t.get_rounds(); ++i) {
        for( int p = 0; p < KpL; ++p)
            buffer[ 2*KpL + p ] = 0;
        for( int j = 0; j < t.get_courts(); ++j) {
            const Match& match = t.get_match( i, j);
            for( int k = 0; k < 4; ++k) {
                int p = match.get_player( k).get_id();
                buffer[ 2*KpL + p ] = 1;
            }
        }
        for( int p = 0; p < KpL; ++p)
            if( buffer[ 2*KpL + p ] == 1) {
                int d = i - buffer[ KpL + p ];
                if( d > buffer[ p ])
                    buffer[ p ] = d;
            } else
                buffer[ KpL + p ] = i;
    }
    
    // Let c indicate the maximum number of subsequent matches of player p.
    for( int p = 0; p < KpL; ++p) {
        int c = buffer[ p ];
        if( c <= 2)
            continue;
        int penalty = weight * (c-2) * (c-2);
        total_penalty += penalty;
        if( dump) {
            std::cout << "player " << Player( p) << " has "
                      << c << " subsequent matches"
                      << " (penalty: " << penalty << ")" << std::endl;
                dumped = true;
            }
    }

    if( dumped)
        std::cout << std::endl;
    return total_penalty;
}

/// Compute penalties for subsequent breaks.
///
/// \param t        The tournament for which to compute the penalty.
/// \param dump     If \c true, dump a message for each penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param weight   The weight for this penalty.
/// \return         The computed penalty.
int penalty_subsequent_breaks( const Tournament& t, int* buffer, bool dump, int weight)
{
    int total_penalty = 0;
    bool dumped = false;
    for( int p = 0; p < KpL; ++p)
        buffer[ p ] = 0;
    for( int p = 0; p < KpL; ++p)
        buffer[ KpL + p ] = -1;

    // For all p in [0,KpL): buffer[ p ] indicates the maximum number of subsequent
    // breaks of player p (so far), buffer[ KpL + p ] indicates the last round
    // where player p had a match.
    for( int i = 0; i < t.get_rounds(); ++i) {
        for( int j = 0; j < t.get_courts(); ++j) {
            const Match& match = t.get_match( i, j);
            for( int k = 0; k < 4; ++k) {
                int p = match.get_player( k).get_id();
                int d = i - buffer[ KpL + p ] - 1;
                if( d > buffer[ p ])
                    buffer[ p ] = d;
                buffer[ KpL + p ] = i;
            }
        }
    }
    for( int p = 0; p < KpL; ++p) {
        int d = t.get_rounds() - buffer[ KpL + p ] - 1;
        if( d > buffer[ p ])
            buffer[ p ] = d;
    }
    
    // Let c indicate the maximum number of subsequent breaks of player p.
    for( int p = 0; p < KpL; ++p) {
        int c = buffer[ p ];
        if( c <= 1)
            continue;
        int penalty = weight * (c-1) * (c-1);
        total_penalty += penalty;
        if( dump) {
            std::cout << "player " << Player( p) << " has "
                      << c << " subsequent breaks"
                      << " (penalty: " << penalty << ")" << std::endl;
                dumped = true;
            }
    }

    if( dumped)
        std::cout << std::endl;
    return total_penalty;
}

/// Evaluate a tournament.
///
/// The value of a tournament is given by the weighted penalities for different
/// constraints.
///
/// \param t        The tournament to evaluate.
/// \param w0       The weight for #penalty_multiple_per_round().
/// \param w1       The weight for #penalty_same_teammate().
/// \param w2       The weight for #penalty_same_opponent().
/// \param w3       The weight for #penalty_subsequent_matches().
/// \param w4       The weight for #penalty_subsequent_breaks().
/// \param dump     If \c true, dump a message for each penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \return         The computed value.
int evaluate( const Tournament& t, int w0, int w1, int w2, int w3, int w4, int* buffer, bool dump)
{
    return penalty_multiple_per_round( t, buffer, dump, w0)
         + penalty_same_teammate( t, buffer, dump, w1)
         + penalty_same_opponent( t, buffer, dump, w2)
         + penalty_subsequent_matches( t, buffer, dump, w3)
         + penalty_subsequent_breaks( t, buffer, dump, w4);
}

/// Perform a random mutation on a tournament.
///
/// The random mutation consists of a swap of two randomly picked players. Both
/// players are either male of female.
///
/// \param t   The tournament to mutate.
/// \return    \c true if the new tournament is valid, \c false otherwise.
bool mutation( Tournament& t)
{
    int i1 = rand() % t.get_rounds();
    int j1 = rand() % t.get_courts();
    Match& m1 = t.get_match( i1, j1);

    int i2 = rand() % t.get_rounds();
    int j2 = rand() % t.get_courts();
    Match& m2 = t.get_match( i2, j2);

    int p1 = 2 * (rand() % 2);
    int p2 = 2 * (rand() % 2);
    if( rand() % 2 == 0) {
        p1 += 1;
        p2 += 1;
    }

    Player p = m1.get_player( p1);
    m1.set_player( p1, m2.get_player( p2));
    m2.set_player( p2, p);    
    
    return m1.is_valid() && m2.is_valid();
}

Tournament run( int w0, int w1, int w2, int w3, int w4, int max_iteration, int* buffer)
{
    Tournament tournament;
    Tournament tmp_tournament;
    Tournament best_tournament;
    int i = 0;
    int value = evaluate( tournament, w0, w1, w2, w3, w4, buffer, false);
    int best_value = value;

    while( i < max_iteration) {
        
        tmp_tournament = tournament;
        bool ok = mutation( tmp_tournament);
        if( !ok) {
            ++i;
            continue;
        }

        int tmp_value = evaluate( tmp_tournament, w0, w1, w2, w3, w4, buffer, false);

        int delta = tmp_value - value;
        if( delta >= 0) {
            if( drand48() <= exp( -V * delta / value)) {
                tournament = tmp_tournament;
                value = tmp_value;
            }
        } else {
            tournament = tmp_tournament;
            value = tmp_value;
            if( value < best_value) {
                best_tournament = tournament;
                best_value = value;
                if( best_value == g_lower_bound)
                    break;
            }
        }

        ++i;
    }
    
    return best_tournament;
}

void usage( const char* name)
{
    std::cout << "Usage: " << name << " <M> <N> <K> <L> <R> <I> [<V> [<w0> <w1> <w2> <w3> <w4>]]" << std::endl;
    std::cout << "    M   number of rounds, M >= 1" << std::endl;
    std::cout << "    N   number of courts, N >= 1" << std::endl;
    std::cout << "    K   number of male players, 2*N <= K <= 26" << std::endl;
    std::cout << "    L   number of female players, 2*N <= N <= 26" << std::endl;
    std::cout << "    R   number of runs, R >= 1 (for example, 100)" << std::endl;
    std::cout << "    I   number of iterations per run, I >= 1 (for example, 10000)" << std::endl;
    std::cout << "    V   volatility of the simulated annealing, V >= 1.0 (default: 30.0)" << std::endl;
    std::cout << "    w0  weight for multiple matcher per round per player, w0 >= 0 (default: 500)" << std::endl;
    std::cout << "    w1  weight for same teammate penalty, w1 >= 0 (default: 100)" << std::endl;
    std::cout << "    w2  weight for same opponent penalty, w2 >= 0 (default: 50)" << std::endl;
    std::cout << "    w3  weight for subsequent matches, w3 >= 0 (default: 5)" << std::endl;
    std::cout << "    w4  weight for subsequent breaks, w4 >= 0 (default: 1)" << std::endl;
}

int main( int argc, char* argv[])
{
    srand( time( NULL));
    
    if( argc != 1+6 && argc != 1+7 && argc != 1+12) {
        usage( argv[0]);
        return 1;
    }
    
    M      = atoi( argv[1]);
    N      = atoi( argv[2]);
    K      = atoi( argv[3]);
    L      = atoi( argv[4]);
    int R  = atoi( argv[5]);
    int I  = atoi( argv[6]);

    V      = argc >= 1+ 7 ? atof( argv[ 7]) :  30;
    int w0 = argc >= 1+12 ? atoi( argv[ 8]) : 500;
    int w1 = argc >= 1+12 ? atoi( argv[ 9]) : 100;
    int w2 = argc >= 1+12 ? atoi( argv[10]) :  50;
    int w3 = argc >= 1+12 ? atoi( argv[11]) :   5;
    int w4 = argc >= 1+12 ? atoi( argv[12]) :   1;

    if( M < 1 || N < 1 || K < 2*N || K > 26 || L < 2*N || L > 26 || R < 1 || I < 1 || V < 1.0f || w0 < 0 || w1 < 0 || w2 < 0 || w3 < 0 || w4 < 0) {
        usage( argv[0]);
        return 2;
    }
        
    KpL = K + L;

    int lb_same_opponent_male   = M*N - K*(K-1)/2;
    int lb_same_opponent_female = M*N - L*(L-1)/2;
    int lb_same_opponent_mixed  = M*N - K*L/2;
    int lb_same_teammate        = 2*M*N - K*L;

    g_lower_bound = 0;
    if( lb_same_opponent_male > 0)
        g_lower_bound += w1 * lb_same_opponent_male;
    if( lb_same_opponent_female > 0)
        g_lower_bound += w1 * lb_same_opponent_female;
    if( lb_same_opponent_mixed > 0)
        g_lower_bound += w1 * lb_same_opponent_mixed;
    if( lb_same_teammate > 0)
        g_lower_bound += w2 * lb_same_teammate;

    Tournament best_tournament;
    int best_value = 1E9;

    Timer timer;
    timer.start();
    
#ifdef HAVE_OPENMP
#pragma omp parallel for schedule(dynamic,1) 
#endif // HAVE_OPENMP
    for( int r = 0; r < R; ++r) {

        if( best_value == g_lower_bound)
            continue;

        int* buffer = new int[ KpL*KpL];

        Tournament tmp_tournament = run( w0, w1, w2, w3, w4, I, buffer);
        int tmp_value = evaluate( tmp_tournament, w0, w1, w2, w3, w4, buffer, false);
        std::cout << "run " << r << " returns value " << tmp_value
                  << ", best value is " << best_value << std::endl;

#ifdef HAVE_OPENMP
#pragma omp critical
#endif // HAVE_OPENMP
        if( tmp_value < best_value) {

            best_tournament = tmp_tournament;
            best_value = tmp_value;

            std::cout << std::endl;
            std::cout << best_tournament;
            std::cout << std::endl;
            evaluate( best_tournament, w0, w1, w2, w3, w4, buffer, true);
        }

        delete[] buffer;

    }
    
    timer.stop();

    // output best result
    std::cout << std::endl;
    std::cout << "best result with value " << best_value << std::endl;
    std::cout << std::endl;
    std::cout << best_tournament;
    std::cout << std::endl;
    
    // output penalties
    int* buffer = new int[ KpL*KpL];
    evaluate( best_tournament, w0, w1, w2, w3, w4, buffer, true);
    delete[] buffer;
    std::cout << std::endl;
    
    // output rounds
    std::vector<std::list<int> > rounds;
    rounds.resize( KpL);
    for( int i = 0; i < best_tournament.get_rounds(); ++i) {
        for( int j = 0; j < best_tournament.get_courts(); ++j) {
            const Match& match = best_tournament.get_match( i, j);
            for( int k = 0; k < 4; ++ k)
                rounds[match.get_player( k).get_id()].push_back( i);
        }
    }
    for( int p = 0; p < KpL; ++ p) {
        std::list<int>::const_iterator it = rounds[p].begin();
        if( it == rounds[p].end()) {
           std::cout << "player " << Player( p) << " plays never" << std::endl;
        } else {
           std::cout << "player " << Player( p) << " plays in round(s) " << (*it++)+1;
           for( ; it != rounds[p].end(); ++it)
               std::cout << ", " << (*it)+1;
           std::cout << " (" << rounds[p].size()
                     << (rounds[p].size() == 1 ? " match)" : " matches)") << std::endl;
        }
    }
    std::cout << std::endl;

    std::cout << "time: " << timer.get_real_time() << " seconds" << std::endl;
    
    return 0;
}
